#include "gapopulation.hh"

#include <algorithm>
#include <iterator>
#include <vector>
#include "pudag.hh"
#include "sched.hh"
#include "verbose.hh"
#include "emit.hh"

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <sched.h>
#ifdef _WPT_
#include <pthread.h>
#include <semaphore.h>
#endif
#include <time.h>
#include <sys/time.h>

#include <iostream>
#include <iomanip>

// Set to something != 0 to quit the calculations. Usually done in sigint..
static volatile sig_atomic_t QUIT_LOOP = 0;

void sig_handler(int sig)
{
  QUIT_LOOP = 1;
}



namespace GA
{


Population::Population(unsigned int _size, Architecture* _architecture, DAG & _dag, Flags & _flags)
  : generation(0), best_individual(0), size(_size), architecture(_architecture), dag(_dag), flags(_flags),
    recreated_schedule_count(0)
{
  matcher = new Matcher(architecture);
  conditions = _architecture->get_conditions();
  
  MUTATE_ALLOW_MULTIPLE = flags.b_ga_allow_multiple_mutations;
  PARENTS_SURVIVE = flags.b_ga_parents_survive;
  STABLE_GENERATIONS_LIMIT = flags.i_ga_stable_stop;
  
  if (flags.d_ga_mutate_node > 100) {
    std::cerr << "(GA) Warning! Node mutation percentage set higher than 100%."
              << " (" << flags.d_ga_mutate_node << ") Using 100%." << std::endl;
    MUTATE_NODE_CHANCE = 100;
  } else if (flags.d_ga_mutate_node < 0) {
    std::cerr << "(GA) Warning! Node mutation percentage set lower than 0%."
              << " (" << flags.d_ga_mutate_node << ") Using 0%." << std::endl;
    MUTATE_NODE_CHANCE = 0;
  } else {
    MUTATE_NODE_CHANCE = flags.d_ga_mutate_node;
  } 
  
  if (flags.d_ga_mutate_inst > 100) {
    std::cerr << "(GA) Warning! Instruction mutation percentage set higher than 100%."
              << " (" << flags.d_ga_mutate_inst << ") Using 100%." << std::endl;
    MUTATE_INST_CHANCE = 100;
  } else if (flags.d_ga_mutate_inst < 0) {
    std::cerr << "(GA) Warning! Instruction mutation percentage set lower than 0%."
              << " (" << flags.d_ga_mutate_inst << ") Using 0%." << std::endl;
    MUTATE_INST_CHANCE = 0;
  } else {
    MUTATE_INST_CHANCE = flags.d_ga_mutate_inst;
  }
  
  if (flags.d_ga_mutate_transfer > 100) {
    std::cerr << "(GA) Warning! Transfer mutation percentage set higher than 100%."
              << " (" << flags.d_ga_mutate_transfer << ") Using 100%." << std::endl;
    MUTATE_TRANSFER_CHANCE = 100;
  } else if (flags.d_ga_mutate_transfer < 0) {
    std::cerr << "(GA) Warning! Transfer mutation percentage set lower than 0%."
              << " (" << flags.d_ga_mutate_transfer << ") Using 0%." << std::endl;
    MUTATE_TRANSFER_CHANCE = 0;
  } else {
    MUTATE_TRANSFER_CHANCE = flags.d_ga_mutate_transfer;
  }
  
  if (flags.d_ga_new_children < 0) {
    std::cerr << "(GA) Warning! New children percentage set lower than 0%."
              << " (" << flags.d_ga_new_children << ") Using 0%." << std::endl;
    NEW_CHILDREN_PERCENTAGE = 0;
  } else {
    NEW_CHILDREN_PERCENTAGE = flags.d_ga_new_children;
  }
  
  if (!PARENTS_SURVIVE && (NEW_CHILDREN_PERCENTAGE < 100)) {
    std::cerr << "(GA) Warning! Parents set to not survive and new children percentage set "
              << "lower than 100% (" << NEW_CHILDREN_PERCENTAGE
              << "). Would cause population size to decrease. Using 100%." << std::endl;
    NEW_CHILDREN_PERCENTAGE = 100;
  }
  
  
  if (flags.b_ga_survivor_truncation && !flags.b_ga_survivor_roulette) {
    survival_method = TRUNCATION_SELECTION;
    //std::cout << "Using truncation for survivor selection." << std::endl;
  } else if (flags.b_ga_survivor_roulette && !flags.b_ga_survivor_truncation) {
    survival_method = FITNESS_PROPORTIONAL_SELECTION;
    //std::cout << "Using roulette wheel for survivor selection." << std::endl;
  } else {
    std::cerr << "(GA) Warning! Confused selection of survivor method. Got either both or "
              << "none (truncation: " << flags.b_ga_survivor_truncation
              << ") (roulette: " << flags.b_ga_survivor_roulette << ")." << std::endl;
    std::cerr << "Using fitness-proportional selection method (roulette wheel)." << std::endl;
    survival_method = FITNESS_PROPORTIONAL_SELECTION;
  }
  
  if (flags.b_ga_bench) {
    BENCH = true;
    //std::cout << "Using bench mode" << std::endl;
  } else {
    BENCH = false;
    //std::cout << "Using normal mode" << std::endl;
  }
  
  if (flags.b_ga_random_search) {
    RANDOM_SEARCH = true;
    PARENTS_SURVIVE = TRUNCATION_SELECTION;
    MUTATE_NODE_CHANCE = 100;
    MUTATE_INST_CHANCE = 100;
    MUTATE_TRANSFER_CHANCE = 100;
    MUTATE_ALLOW_MULTIPLE = false;
    std::cout << "Using random search" << std::endl;
  } else {
    RANDOM_SEARCH = false;
  }
  if (flags.b_ga_pthreads) {
#ifndef _WPT_
    std::cerr << "You must compile optimist with -D_WPT_ to use -gapthreads" 
	      << std::endl;
    exit(-1);
#else
    std::cout << "Using pthreads for creating schedules" << size << std::endl;
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    /* Set up args with semaphores */
    std::cout << "Setting up threads " << flags.i_ga_size << " ... ";
    child = new pthread_t[flags.i_ga_size+2];
    arg = new thread_create_schedule_arg[flags.i_ga_size+2];
    
    for(int i=0; i<flags.i_ga_size+2; i++){
      arg[i].nr = i;
      arg[i].start = new sem_t;
      sem_init(arg[i].start, 0, 0);
      arg[i].done = new sem_t;
      sem_init(arg[i].done, 0, 0);
      pthread_create(&child[i], &attr, thread_create_schedule, (void *)&arg[i]);
      std::cout << i << ", ";
    }
    std::cout << std::endl;
    pthread_attr_destroy(&attr);
#endif
  }
}

Population::~Population()
{
  delete_individuals(individuals);
  delete best_individual;
  best_individual = 0;
  delete matcher;
  matcher = 0;
  if(flags.b_ga_pthreads) {
#ifdef _WPT_
    for(int i=0; i<flags.i_ga_size+2; i++) {
      delete arg[i].start;
      delete arg[i].done;
    }
    delete[] child;
    delete[] arg;
#endif
  }
}

void Population::delete_individuals(individuals_t& _individuals)
{
  // Loop through individuals and delete them and their schedules.
  individuals_t::iterator it;
  for (it = _individuals.begin(); it != _individuals.end(); ++it) {
    /* Does not remove it from the STL container, just deletes the
       object, so the iterator is still valid. */
    delete *it;
    *it = 0;
  }
  _individuals.clear();
}

/**
 * Initializes the population by generating 'size' new random
 * individuals. Used to create a new initial population. May be used
 * to re-create a new initial population.
 */
void Population::initialize()
{
  // Delete all existing individuals.
  delete_individuals(individuals);
  if (best_individual != NULL) {
    delete best_individual;
  }
  
  generation = 0;
  recreated_schedule_count = 0;
  
  if (RANDOM_SEARCH) {
    size = 1;
  }
  
  // Handle Ctrl-C.
  struct sigaction new_action, old_action;
  memset(&new_action, 0, sizeof(new_action));
  new_action.sa_handler = sig_handler;
  sigemptyset(&new_action.sa_mask);
  new_action.sa_flags = 0;
  sigaction(SIGINT, &new_action, NULL);
  sigaction(SIGXCPU, &new_action, NULL); // Good with "ulimit -S -t" (soft cpu time limit).
  sigaction(SIGTERM, &new_action, NULL); // Good with "kill <PID>" when you want the result.
  
  if (!BENCH) {
    std::cout << "Creating initial Individuals..." << std::endl;
  }
  for (unsigned int i=0; i < size; ++i) {
    if (QUIT_LOOP != 0) {
      std::cerr << "SIGINT, SIGXCPU or SIGTERM caught. Interrupting initialization." << std::endl;
      break;
    }
    Individual* ind = generate_individual();
    if (ind != 0) {
      individuals.push_back(ind);
    }
  }
  
  if (!BENCH) {
    std::cout << individuals.size() << " individuals created.\n";
  }
  
  calculate_fitness(individuals);
  std::sort(individuals.begin(), individuals.end(), better_individual_first());
  
  if (individuals.size() > 0) {
    best_individual = new Individual(*individuals.front());
    std::cout << "Best at start: " << get_best_individual()->get_fitness() << "\n";
  }
  
  if (QUIT_LOOP != 0) {
    std::cout << "Recreated schedules: " << recreated_schedule_count << std::endl;
    std::cout << std::flush;
    exit(0);
  }
  
  // Restore the handling of SIGINT and SIGXCPU. We no longer want to handle it ourself.
  sigaction(SIGXCPU, &old_action, NULL);
  sigaction(SIGINT, &old_action, NULL);
  sigaction(SIGTERM, &old_action, NULL);
}

void Population::calculate_fitness(individuals_t& _individuals)
{
  individuals_t::iterator it = _individuals.begin();
  for (; it != _individuals.end(); ++it) {
    (*it)->calculate_fitness();
  }
}


/* slaves */
#define ping {std::cout << "*ping*" << nr << ", " << __LINE__ << "\n";}
void * thread_create_schedule( void * arg )
{
#ifdef _WPT_
  int nr =  ((thread_create_schedule_arg*)arg)->nr;  
  std::cout << "thread " << nr << " starting" << std::endl;
  while(1){
    sem_wait(((thread_create_schedule_arg*)arg)->start);
    Individual ** ind = ((thread_create_schedule_arg*)arg)->ind;
    Population * pop = ((thread_create_schedule_arg*)arg)->pop;

    if(NULL == pop->create_schedule(*ind, false, false, true)){
      // Couldn't create a schedule. Generating a new individual.
      std::cout << "Re-creating an Individual." << std::endl;
      delete *ind;
      *ind = pop->generate_individual();
      pop->inc_recreated_schedule_count(); // The first try isn't
                                           // counted in
                                           // generate_individual().
    }
    sem_post(((thread_create_schedule_arg*)arg)->done);
  }
  pthread_exit(0);
#endif
}

/* mater: candidate for parallelism, around 80% of execution time
   spent here for a few test cases */

void Population::create_schedules(individuals_t& _individuals)
{ 
  if(flags.b_ga_pthreads){
#ifdef _WPT_
    int i,j;
    /* MSTEP is useless for anything bug debugging */
    /* Set it to 1 for normal parallel execution */
#define MSTEP 1
    i=j=0;
    for(int steps=0; steps<MSTEP; steps++){
      for(; i < _individuals.size()/MSTEP*(steps+1); i++){
	arg[i].ind = &(_individuals[i]);
	arg[i].pop = this;
	sem_post(arg[i].start);
      }
      
      for(; j < _individuals.size()/MSTEP*(steps+1); j++){
	sem_wait(arg[j].done);
      }
    }
    for(; i < _individuals.size(); i++){
      arg[i].ind = &(_individuals[i]);
      arg[i].pop = this;
      sem_post(arg[i].start);
    }

    for(; j < _individuals.size(); j++){
      sem_wait(arg[j].done);
    }
#endif
  }
  else{
//     individuals_t::reverse_iterator it0 = _individuals.rbegin();
//     for (; it0 != _individuals.rend(); ++it0) {
//       gnodes_t::iterator git = (*it0)->get_nodes().begin();
//       for (; git != (*it0)->get_nodes().end(); ++git) {
// 	delete *git;
// 	*git = 0;
//       }
//       (*it0)->get_nodes().clear();
//     }

    individuals_t::reverse_iterator it = _individuals.rbegin();
    struct timeval t0, t1;
    struct timezone d0, d1;
    int i;
    std::cout << "Doing " << _individuals.size() << " inds\n";
    for (i=0; it != _individuals.rend(); ++it, ++i) {
      gettimeofday(&t0, &d0);
      if (create_schedule(*it) == NULL) {
	// Couldn't create a schedule. Generating a new individual.
	//std::cout << "Re-creating an Individual." << std::endl;
	delete *it;
	*it = generate_individual();
	++recreated_schedule_count; // The first try isn't counted in generate_individual().
      }
      gettimeofday(&t1, &d0);
      std::cout << std::endl << i << ":" << ((t1.tv_sec - t0.tv_sec)*1000000 + t1.tv_usec-t0.tv_usec)/1000 << std::endl;
    }
  }
}

/**
 * Evolves the population max_times generations.
 */
void Population::evolve(unsigned int max_times)
{
  if (individuals.empty()) {
    std::cout << "No evolution performed. The population needs to be of size > 0" << std::endl;
    return;
  }
  
  if (!BENCH) {
    std::cout << "Starting evolution.\n";
  }
  
  // Handle Ctrl-C.
  struct sigaction new_action, old_action;
  memset(&new_action, 0, sizeof(new_action));
  new_action.sa_handler = sig_handler;
  sigemptyset(&new_action.sa_mask);
  new_action.sa_flags = 0;
  sigaction(SIGINT, &new_action, NULL);
  sigaction(SIGXCPU, &new_action, NULL); // Good with "ulimit -S -t" (soft cpu time limit).
  sigaction(SIGTERM, &new_action, NULL); // Good with "kill <PID>" when you want the result.
  
  unsigned int loops = 0;
  int num_stable_loops = 1;
  fitness_t last_fitness = -1;
  
  while (loops < max_times) {
    if (QUIT_LOOP != 0) {
      std::cerr << "SIGINT, SIGXCPU or SIGTERM caught. Interrupting evolution." << std::endl;
      break;
    }
    
    fitness_t fitness = advance(); // Advance one generation.
    if (fitness <= best_individual->get_fitness()) {
      delete best_individual;
      best_individual = new Individual(*individuals.front()); // individuals should be in sorted order
    }
    
    if (last_fitness == fitness) {
      /* This more or less assumes that we deal with integer fitness
         values. This is another reason to introduce a Fitness class
         and fix this. For real value fitness this should be (psuedo)
         last_fitness == (fitness +- epsilon) */
      ++num_stable_loops;
    } else {
      num_stable_loops = 1;
      last_fitness = fitness;
    }
    
    std::cout << "Loop " << loops << ", Individuals: ";
    for (unsigned int i=0; i < individuals.size(); ++i) {
      std::cout << individuals[i]->get_fitness()
	//<< "(" << individuals[i]->get_birth_date() << ")"
                << ", ";
    }
    std::cout << std::endl;
    //std::cout << "Number of parents: " << parents.size() << std::endl;
    
    ++loops;
    
    // Abort if we have had a certain number of generations with stable best fitness.
    if ((STABLE_GENERATIONS_LIMIT > 0) && (num_stable_loops >= STABLE_GENERATIONS_LIMIT)) {
      if (!BENCH) {
        std::cout << "The fitness was stable for " << num_stable_loops
                  << " generations. Stopping evolution." << std::endl;
      } else {
        std::cout << "Stable for loops: " << num_stable_loops << std::endl;
      }
      break;
    }
  }
  
  std::cout << "Recreated schedules: " << recreated_schedule_count << std::endl;
  //  std::cout << "The best individual is:\n" << *get_best_individual();
  //  std::cout << "\n----------------------\n";

  if (BENCH) {
    std::cout << "Generations: " << loops << std::endl;
    std::cout << "Best at end: " << get_best_individual()->get_fitness() << std::endl;
  }
  
  // Restore the handling of SIGINT and SIGXCPU. We no longer want to handle it ourself.
  sigaction(SIGXCPU, &old_action, NULL);
  sigaction(SIGINT, &old_action, NULL);
  sigaction(SIGTERM, &old_action, NULL);
}

/**
 * Advances the population one generation.
 * Returns the fitness of the best individual in that generation.
 */
fitness_t Population::advance()
{
  individuals_t new_individuals;
  
  if (RANDOM_SEARCH) {
    Individual* new_ind = generate_individual();
    new_individuals.push_back(new_ind);
    
  } else {
    // Create new children with crossover.
    while (new_individuals.size() < (size * (NEW_CHILDREN_PERCENTAGE / 100.0))) {
      /* TODO: Add another limit to the loop, like a counter? Could
         cause decreasing size, but could also save us from an
         infinite loop. */
      std::pair<GA::Individual*, GA::Individual*> childs =
        crossover(select_parent(individuals), select_parent(individuals));
      
      if (childs.first != 0 && childs.second != 0) {
        // We got two new valid Inviduals.
        new_individuals.push_back(childs.first);
        new_individuals.push_back(childs.second);
      }
    }
    
    if (!PARENTS_SURVIVE) {
      delete_individuals(individuals);
    }
  
    mutate(new_individuals);
    create_schedules(new_individuals);
  }
  
  calculate_fitness(new_individuals);
  individuals.insert(individuals.end(), new_individuals.begin(), new_individuals.end());
  new_individuals.clear(); // The new individuals were copied to individuals.
  
  // All 'individuals' + 'new_individuals' should compete for the 'size' spots.
  std::sort(individuals.begin(), individuals.end(), better_individual_first());
  
  if (individuals.size() > size) {
    select_survivors(individuals, size);
  }
  
  ++generation;
  
  if (individuals.size() > 0) {
    // Since we sorted the population earlier, the first individual is the one with highest fitness.
    return individuals.front()->get_fitness();
  } else {
    return fitness_t(-1.0);
  }
}

void Population::mutate(individuals_t& _individuals)
{
  // Mutate children.
  if (MUTATE_ALLOW_MULTIPLE) {
    for (unsigned int i=0; i < _individuals.size(); ++i) {
      if (drandom_interval(0, 100) <= MUTATE_NODE_CHANCE) {
        mutate_node(_individuals[i]);
      }
      if (drandom_interval(0, 100) <= MUTATE_INST_CHANCE) {
        mutate_inst(_individuals[i]);
      }
#ifdef _SP_
#ifdef _TRANS_
      if (drandom_interval(0, 100) <= MUTATE_TRANSFER_CHANCE) {
        mutate_transfer(_individuals[i]);
      }
#endif
#endif
    }
  } else {
    for (unsigned int i=0; i < _individuals.size(); ++i) {
      // TODO: Fix this? The first kind of mutation have higher chance
      // to occur than the second.
      bool mutated = false;
      if (!mutated && (drandom_interval(0, 100) <= MUTATE_NODE_CHANCE)) {
        mutate_node(_individuals[i]);
        mutated = true;
      }
      if (!mutated && (drandom_interval(0, 100) <= MUTATE_INST_CHANCE)) {
        mutate_inst(_individuals[i]);
        mutated = true;
      }
#ifdef _SP_
#ifdef _TRANS_
      if (!mutated && (drandom_interval(0, 100) <= MUTATE_TRANSFER_CHANCE)) {
        mutate_transfer(_individuals[i]);
        mutated = true;
      }
#endif
#endif
    }
  }
}

Individual* Population::select_parent(individuals_t& _individuals)
{
  // TODO: Support more than binary tournament? Some way to select which kind.
  if (_individuals.empty()) {
    std::cerr << "Can't select any parent. No parents exists." << std::endl;
    exit(-1);
  }
  
  return binary_tournament_selection(_individuals);
}

Individual* Population::binary_tournament_selection(individuals_t& _individuals)
{
  int p1 = random_interval(0, _individuals.size());
  int p2 = random_interval(0, _individuals.size());
  
  if (_individuals[p1]->get_fitness() <= _individuals[p2]->get_fitness()) {
    return _individuals[p1];
  } else {
    return _individuals[p2];
  }
}

/**
 * Selects 'num' individuals from 'individuals' using a
 * fitness-proportional selection method, the so called "roulette
 * wheel" selection. The selected individuals are kept in
 * 'individuals' and the unselected individuals are deleted. If 'num'
 * equals the number of individuals supplied, nothing is done, because
 * all would be selected.
 * 
 * Assumes that 'individuals' is sorted in descending order of
 * fitness, and that lower fitness value is better. Also assumes that
 * there are enough individuals to select 'num' unique ones, if not,
 * nothing is done.
 */
void Population::roulette_wheel_selection(individuals_t& individuals, unsigned int num)
{
  individuals_t survivors;
  
  if (num >= individuals.size()) {
    return;
  }
  
  std::vector<double> modified_fitness;
  modified_fitness.reserve(individuals.size());
  
  // std::cout << "--- roulett wheel ---" << std::endl;
  
  //std::cout << "Size: " << individuals.size() << std::endl;
  
  // min_size is added to max_fitness so that we don't get 0 as modified fitness value on the worst individual.
  double min_size = 1;
  double max_fitness = individuals.back()->get_fitness().get_value() + min_size;
  double total_fitness = 0;
  
  //std::cout << "Max: " << max_fitness << std::endl;
  
  for (unsigned int i=0; i < individuals.size(); ++i) {
    // We want to have a fitness value where higher means better.
    modified_fitness.push_back(max_fitness - individuals[i]->get_fitness().get_value());
    total_fitness += modified_fitness.back();
    
    //std::cout << "  " << i << ": " << individuals[i]->get_fitness().get_value() << ", modified: "
    //          << modified_fitness.back() << std::endl << ", total: " << total_fitness << std::endl;
  }
  
  //std::cout << "Total: " << total_fitness << std::endl;
  
  while (survivors.size() < num) {
    double n = drandom_interval(0, total_fitness);
    //std::cout << "n: " << n << std::endl;
    //std::cout << "sum: ";
    double sum = 0;
    for (unsigned int k=0; k < individuals.size(); ++k) {
      sum += modified_fitness[k];
      //std::cout << sum << ", ";
      if (sum >= n && individuals[k] != 0) {
        //std::cout << std::endl;
        //std::cout << "Fitness: " << individuals[k]->get_fitness() << std::endl;
        survivors.push_back(individuals[k]);
        individuals[k] = 0; // Skip from deleting.
        break;
      }
    }
  }
  
  individuals_t::iterator it = individuals.begin();
  for (; it != individuals.end(); ++it) {
    if ((*it) != 0) {
      delete *it;
      *it = 0;
    }
  }
  
  individuals = survivors;
  std::sort(individuals.begin(), individuals.end(), better_individual_first());
  
  //std::cout << "--- end of roulett wheel ---" << std::endl;
}


/**
 * Selects 'num' individuals from 'individuals' by trimming the size
 * of the population. It removes the worst excessive Individuals. It
 * can be very aggressive in terms of takeover speed.
 */
void Population::truncation_selection(individuals_t& individuals, unsigned int num)
{
  unsigned int SAVE_NUM_WORST = 0;
  
  int diff = individuals.size() - num;
  unsigned int m = SAVE_NUM_WORST;
  if (m > num) {
    m = num;
  }
  
  individuals_t::iterator d_end = individuals.end() - m;
  individuals_t::iterator d_begin = d_end - diff;
  for (individuals_t::iterator it=d_begin; it < d_end; ++it) {
    delete *it;
    *it = 0;
  }
  individuals.erase(d_begin, d_end);
}

/**
 * Selects 'num' individuals in 'individuals' that survives. Excess
 * individuals are deleted.
 */
void Population::select_survivors(individuals_t& individuals, unsigned int num)
{
  // Call roulette_wheel_selection as standard.
  // How to implement possiblity to select between roulette and trim - how to implement trim in the same way?

  switch (survival_method) {
  case TRUNCATION_SELECTION: {
    truncation_selection(individuals, num);
    break;
  }
  case FITNESS_PROPORTIONAL_SELECTION:
  default: {
    roulette_wheel_selection(individuals, num);
    break;
  }
  }
}

/*
 * Returns a pointer to the best Individual that was created the last
 * evolution. This is only valid after a population has been
 * initialized and there is at least one individual in the population.
 * Always check that it's not 0 before using it.
 */
Individual* Population::get_best_individual()
{
  return best_individual;
}

/**
 * Tries to find a common crossover point for the two given
 * parents. Randomly selects one of the points found, if more than
 * one. Creates a pair of new child Individuals if a point was
 * found. Both parents must have a valid data. If a mutation has been
 * done, then a new Schedule must be built before it can be used as a
 * parent for crossover.
 * 
 * Returns a pair of new Individuals if a crossover point could be
 * found. Otherwise a pair<0, 0>.
 */
std::pair<Individual*, Individual*> Population::crossover(Individual* parent1, Individual* parent2)
{
  /*We skip the first and the last node, because they aren't
    interesting for crossover. The first node will always be identical
    and the last node is not interesting because if it's identical,
    we will get two childs that look exactly as the parents.
  */
  if (parent1 == parent2) {
    // TODO: Do we want to allow the same Individual used as both parents?
    //std::cerr << "Warning! Crossover got the same Individual as both parents.\n";
    //return std::make_pair(0, 0);
  }
  
  Individual* child1 = 0;
  Individual* child2 = 0;
  unsigned int p1_size = parent1->get_nodes().size();
  unsigned int p2_size = parent2->get_nodes().size();
  std::vector< std::pair<unsigned int, unsigned int> > matches;

  /* Find matching */
  for (unsigned int n = 1; n < p1_size - 1; ++n) {
    GNode* gn1 = parent1->get_nodes()[n];
    for (unsigned int m = 1; m < p2_size - 1; ++m) {
      GNode* gn2 = parent2->get_nodes()[m];
      
      if ( (gn1->get_z() == gn2->get_z())
           && (*gn1->get_time_profile() == *gn2->get_time_profile())
#ifdef _SP_
           && (*gn1->get_space_profile() == *gn2->get_space_profile())
#endif //_SP_
           ) {
        matches.push_back(std::make_pair(n, m));
        //std::cout << "(crossover) Match on node: " << n << "\n";
      }
    }
  }
  
  
  if (matches.size() > 0) {
    // Pick a random match.
    std::pair<unsigned int, unsigned int> match = matches[random_interval(0, matches.size())];
    unsigned int p1_cross = match.first;
    unsigned int p2_cross = match.second;
    
//     std::cout << "(crossover) Found " << matches.size() << " matching nodes for crossover. "
//               << "Selected node " << p1_cross << " and " << p2_cross  << ".\n";
    //std::cout << *parent1 << *parent2; // This would cause weird problems if VERB was used! (or?)
    

    child1 = new Individual();
    child2 = new Individual();
    
    child1->set_birth_date(generation);
    child2->set_birth_date(generation);
    
    // Create new GNode copies for the children.
    for (unsigned int i=0; i < p1_cross; ++i) {
      child1->get_nodes().push_back(new GNode(*(parent1->get_nodes()[i])));
    }
    for (unsigned int i=0; i < p2_cross; ++i) {
      child2->get_nodes().push_back(new GNode(*(parent2->get_nodes()[i])));
    }
    
    for (unsigned int j=p2_cross; j < p2_size; ++j) {
      child1->get_nodes().push_back(new GNode(*(parent2->get_nodes()[j])));
    }
    for (unsigned int j=p1_cross; j < p1_size; ++j) {
      child2->get_nodes().push_back(new GNode(*(parent1->get_nodes()[j])));
    }
    
    //std::cout << "building zorder" << std::endl;
    build_zorder(child1);
    build_zorder(child2);
    
    //std::cout << "building prios" << std::endl;
    build_instruction_priorities(child1);
    build_instruction_priorities(child2);
    
#ifdef _SP_
#ifdef _TRANS_
    build_transfer_priorities(child1);
    build_transfer_priorities(child2);
#endif
#endif
  }
  
  return std::make_pair(child1, child2);
}

/**
 * Mutates the Individual by randomly selecting a GNode and making a
 * different choice from its zero-indegree set. It's required to
 * create a new Schedule for this change to take full effect. The
 * Individual is not correct until that has been done. This function
 * has a limited amount (MAX_TRIES) of times it tries to find a GNode
 * with more than 1 node in its z-set.
 */
void Population::mutate_node(Individual* ind)
{
  //std::cout << "(GA::mutate_node)" << std::endl;
  
  int size = ind->get_nodes().size();
  bool found_matching = false;
  int MAX_TRIES = 5; // Maximum number of tries to randomly select a node.
  int tries = 0; // How many times we have tried finding a node.
  zorder_t& zorder = ind->get_z_order();
  vertex_id new_z_choice;
  
  GNode* gn1;
  int n1;
  
  while (!found_matching && tries < MAX_TRIES) {
    n1 = random_interval(0, size-1);
    gn1 = ind->get_nodes()[n1];
    vertex_set_t& z = gn1->get_z();
    if (z.size() > 1) {
      // There are more than one possible node to select from z.
      std::vector<vertex_id> tmpl;
      tmpl.push_back(gn1->get_z_choice());
      std::vector<vertex_id> diff;
      std::insert_iterator<std::vector<vertex_id> > ii(diff, diff.begin());
      std::set_difference(z.begin(), z.end(), tmpl.begin(), tmpl.end(), ii);
      
      if (diff.size() > 0) {
        new_z_choice = diff[random_interval(0, diff.size())];
        found_matching = true;
      } else {
        /* This should never happen, since z.size() > 1 and we only remove
           one element (the last z_choice). */
        std::cerr << "(GA::mutate) Error: No elements to choose from, but z.size() > 1!" << std::endl;
        exit(-1);
      }
    }
    ++tries;
  }
  
  if (found_matching) {
    // Make the changes in zorder.
    vertex_id old_z_choice = gn1->get_z_choice();
//     std::cout << "Choosing " << new_z_choice << " instead of " << old_z_choice
//               << " at node " << n1 << "\n";
    
    // "Move" new_z_choice to the position before old_z_choice.
    zorder.remove(new_z_choice);
    zorder_t::iterator it = std::find(zorder.begin(), zorder.end(), old_z_choice);
    zorder.insert(it, new_z_choice);
  } else {
    //std::cout << "Didn't find any node to mutate.\n";
  }
}

/**
 * Mutates the Individual by randomly removing an instruction
 * priority. It's required to create a new Schedule for this change to
 * take full effect. The Individual is not correct until that has been
 * done. When the new Schedule is created, it will randomly select an
 * instruction from the candidates, instead of using the prioritized
 * one.
 */
void Population::mutate_inst(Individual* ind)
{
  //std::cout << "(GA::mutate_inst)" << std::endl;
  
  // The last GNode's z_choice never has an instruction priority.
  int n = random_interval(0, ind->get_nodes().size() - 1);
  const vertex_id& foo = ind->get_nodes()[n]->get_z_choice();
  ind->get_instruction_priorities().erase(foo);
  /* New instruction priorities are set when create_schedule() is run,
     or if build_instruction_priorities() is called directly. */
}

/**
 * TODO
 */
void Population::mutate_transfer(Individual* ind)
{
  //std::cout << "(GA::mutate_transfer)" << std::endl;
  
#ifdef _SP_
#ifdef _TRANS_
  std::vector<std::vector<TargetInstruction*>::iterator> instructions;
  
  transfer_priority_t& transprio = ind->get_transfer_priorities();
  transfer_priority_t::iterator transit = transprio.begin();
  for (; transit != transprio.end(); ++transit) {
    std::vector<TargetInstruction*>::iterator instit = (*transit).second.begin();
    for (; instit != (*transit).second.end(); ++instit) {
      if (*instit != 0) {
        instructions.push_back(instit);
      }
    }
  }
  
  if (instructions.size() > 0) {
    // Select a random priority and set it to 0 (equal to removing it).
    int n = random_interval(0, instructions.size());
    *instructions[n] = 0;
  }
#endif
#endif
}

/**
 * TODO
 */
void Population::build_transfer_priorities(Individual* ind)
{
#ifdef _SP_
#ifdef _TRANS_
  //std::cout << "Building transfer priorities (for " << ind << ")" << std::endl;
  transfer_priority_t& transprio = ind->get_transfer_priorities();
  transprio.clear();
  
  gnodes_t::iterator it = ind->get_nodes().begin() + 1; // The first node doesn't have an instruction
  gnodes_t::iterator end = ind->get_nodes().end();
  GNode* last_normal = ind->get_nodes().front();
  for (; it != end; ++it) {
    //std::cout << "last_normal: " << last_normal->get_z_choice() << std::endl;
    if ((*it)->get_instruction_type() == TRANSFER_INSTRUCTION) {
      vertex_id u_id = last_normal->get_z_choice();
      transfer_priority_t::iterator ins_it =
        transprio.insert(std::make_pair(u_id, std::vector<TargetInstruction*>())).first;
      (*ins_it).second.push_back((*it)->get_target_instruction());
      //std::cout << "Added transfer priority for " << u_id << " + " << (*ins_it).second.size()-1
      //          << " for instruction " << (*it)->get_target_instruction()->get_mnemonic()
      //          << std::endl;
    }
    if ((*it)->get_z_choice() != boost::graph_traits<DAG>::null_vertex()) {
      last_normal = *it;
    }
  }
#endif
#endif
}

/**
 * Builds a new instruction priority map for an Individual, based on
 * the choices stored in its GNodes.
 */
void Population::build_instruction_priorities(Individual* ind)
{
//   std::cout << "Building instruction priorities (for " << ind << ")" << std::endl;
  instruction_priority_t& instprio = ind->get_instruction_priorities();
  instprio.clear();
  
  /* We want to save the choice made in one GNode and the resulting
     instruction from the node we ended up in by making that choice. */
  gnodes_t::iterator it = ind->get_nodes().begin() + 1; // The first node doesn't have an instruction
  gnodes_t::iterator end = ind->get_nodes().end();
  for (; it != end; ++it) {
    vertex_id u_id = (*(it - 1))->get_z_choice(); // The choice was made in the previous node.
    if (u_id != boost::graph_traits<DAG>::null_vertex()) {
      instprio.insert(std::make_pair(u_id, (*it)->get_target_instruction()));
//       std::cout << "  Added instruction priority for " << u_id << " to "
//                 << (*it)->get_target_instruction()->get_mnemonic() << std::endl;
    }
  }
}

/**
 * Builds a new zorder for an Individual, based on the choices stored
 * in its GNodes.
 */
void Population::build_zorder(Individual* ind)
{
  zorder_t& zorder = ind->get_z_order();
  zorder.clear();
  
  gnodes_t::iterator it = ind->get_nodes().begin();
  for (; it != ind->get_nodes().end(); ++it) {
    if ((*it)->get_z_choice() != boost::graph_traits<DAG>::null_vertex()) {
      zorder.push_back((*it)->get_z_choice());
    }
  }
}

/**
 * Find the first element in zorder that exist in z too.
 * 
 * TODO: There's a possiblity that we always choose elements in order
 * from zorder (a and thus this function is unnessecery). See
 * create_schedule(). Edit: Reconsider this with transfers. Changes anything?
 */
zorder_t::iterator Population::get_first_z_element_in_zorder(const vertex_set_t& z, zorder_t& zorder)
{
  zorder_t::iterator zorder_it = zorder.begin();
  vertex_set_t::const_iterator z_find;
  for (; zorder_it != zorder.end(); ++zorder_it) {
    z_find = std::find(z.begin(), z.end(), *zorder_it);
    if (z_find != z.end()) {
      return zorder_it;
    }
  }
  return zorder.end();
}

/**
 * Generate a new random individual.
 */
Individual* Population::generate_individual()
{
  Individual* ind = new Individual();
  Schedule* s = create_schedule(ind, true);
  bool tried_simple = false;
  
  while (s == NULL) {
    delete ind;
    
    if (QUIT_LOOP != 0) {
      std::cerr << "SIGINT, SIGXCPU or SIGTERM caught. "
                << "Interrupting attempts to generate an Individual." << std::endl;
      return NULL;
    }
    
    ind = new Individual();
    //std::cout << "Re-creating a schedule..." << std::endl;
    if (false == tried_simple) {
      // Try doing a simple.
      // We only try the simple version once, because it will be the same every time.
      s = create_schedule(ind, true, true);
      tried_simple = true;
    } else {
      s = create_schedule(ind, true);
    }
    ++recreated_schedule_count;
  }
  
  ind->set_birth_date(generation);
  return ind;
}

/**
 * Creates a new Schedule for the Individual using the data stored
 * in the GNodes. Deletes the old Schedule in the Individual, stores
 * the new Schedule in the Individual and returns a pointer to it.
 * 
 * If new_individual is set, it makes random choices instead of using
 * the information stored in the Individuals GNodes, mostly used when
 * generating a Individual, which doesn't yet have any GNodes.
 * 
 * [S]imple_heuristics means that it always takes the first node in the
 * z-set, instead of taking a random. This is to mimic the hs=1 option
 * to Optimist.
 */
#define nping {std::cout << "*ping*, " << pthread_self() << ", " <<  __LINE__ << "\n";}
#define lockit {if(parallel) pthread_mutex_lock(&mutx);}
#define unlockit {if(parallel) pthread_mutex_unlock(&mutx);}

/* Nice timing macro */
#define delta                \
 {                           \
  gettimeofday(&et, NULL);   \
  std::cout << std::setw(5) << ((et.tv_sec - st.tv_sec)*1000000 + et.tv_usec-st.tv_usec)/1000 << ", ";\
  st = et;                   \
 }
#undef delta
#define delta {}

Schedule* Population::create_schedule(Individual* ind, bool new_individual, bool simple_heuristics, bool parallel)
{
  //std::cout << "--- (ind: " << ind << ")" << std::endl << *ind << std::endl << "---" << std::endl;
  //std::cout << "--- Create Schedule --- (New? " << new_individual << ")" << std::endl;
  struct timeval st, et;
  gettimeofday(&st, NULL);

  gnodes_t::iterator git = ind->get_nodes().begin();
  for (; git != ind->get_nodes().end(); ++git) {
    delete *git;
    *git = 0;
  }
  ind->get_nodes().clear();
  delta; /*1*/
  
  // We must copy zorder, because we modify it temporary in this function.
  zorder_t zorder = ind->get_z_order();
  gnodes_t& new_gnodes = ind->get_nodes();
  unsigned int num_transfers = 0; // Number of transfers since last selected vertex_id.
  vertex_id last_selected = boost::graph_traits<DAG>::null_vertex();
  delta; /*2*/

  ZSet * zset = new ZSet();
  std::list< SNode* > snodes;
  
  NodeAttr* node_attr = this->get_init_node_attr(dag, zset);
  TimeProfile* init_time_profile = new TimeProfile(architecture->get_reserv_tab_size());
  delta; /*3*/
  SNode* init_snode = new SNode(node_attr, init_time_profile, architecture);

  delta;/*4*/

  delete init_time_profile;

  snodes.push_back(init_snode);
    
  GNode* init_gnode = new GNode();
  init_gnode->set_z(init_snode->get_node_attr()->z);
  init_gnode->set_reference_time(init_snode->get_reference_time());
  init_gnode->set_tau(init_snode->get_tau());
  init_gnode->set_time_profile(new TimeProfile(*init_snode->get_time_profile()));
#ifdef _SP_
  init_gnode->set_space_profile(new SpaceProfile(*init_snode->get_space_profile()));
#endif  
  init_gnode->set_instruction_type(NO_INSTRUCTION);
  new_gnodes.push_back(init_gnode);
  delta; /*5*/
  vertex_set_t z = init_snode->get_node_attr()->z;
  int cnt=0;
  while (!z.empty()) {
    cnt++;
    SNode* current_snode = snodes.back();
    instruction_candidates_t candidates;
    vertex_id selected = boost::graph_traits<DAG>::null_vertex();
    zorder_t::iterator zorderit = zorder.end();
    
    if (new_individual) {
      vertex_set_t::iterator zit = z.begin();
      
      if (!simple_heuristics) {
        int n = random_interval(0, z.size());
        
        // TEST: Comment out for always selecting the first node.
        for (int i = 0; i < n; ++i) {
          ++zit;
        }
      }
      selected = *zit;
    } else {
      zorderit = get_first_z_element_in_zorder(z, zorder);
      
      if (zorderit != zorder.end()) {
        selected = *zorderit;
      } else {
        /* zorder doesn't contain any of z's nodes. We have a new node
         to select. This happens if you previously had a pattern that
         covered more than one node and now the node that wasn't
         selected is the only one left in z.*/
        vertex_set_t::iterator zit = z.begin();
        
        if (!simple_heuristics) {
          int n = random_interval(0, z.size());
          
          for (int i = 0; i < n; ++i) {
            ++zit;
          }
        }
        selected = *zit;
      }
    }

    GNode* gnode = new GNode();

    match_node(candidates, selected, current_snode, dag, snodes, gnode, ind);


    transfers(candidates, current_snode, snodes, gnode, ind);
    InstructionType type =
      select_instruction(candidates, selected, last_selected, current_snode,
                         snodes, gnode, ind, num_transfers, zset);

    if (type == NORMAL_INSTRUCTION) {
      num_transfers = 0;
      new_gnodes.back()->set_z_choice(selected);
      last_selected = selected;
      if (zorderit != zorder.end()) { // We actually selected a node in zorder.
        /* Idea: If we only use each element one time, so we can
           remove them and gain speed in the next search. */
        zorder.erase(zorderit);
      }
#ifdef _SP_
#ifdef _TRANS_
    } else if (type == TRANSFER_INSTRUCTION) {
      ++num_transfers;
      new_gnodes.back()->set_z_choice(boost::graph_traits<DAG>::null_vertex());
#endif
#endif
    } else if (type == NO_INSTRUCTION) {
      // We couldn't select any instruction. Throw away this schedule.
      delete gnode;
      std::list<SNode*>::iterator sit = snodes.begin();
      for (; sit != snodes.end(); ++sit) {
        delete *sit;
        *sit = 0;
      }
      
      delete zset;
      zset = 0;
      
      return 0;
    } else {
      std::cerr << "(GA::create_schedule) Error: Unknown instruction type!" << std::endl;
      exit(-1);
    }
    
    candidates.clear();
    new_gnodes.push_back(gnode);
    
    // We have added the successor SNode in update_space. Get the next z-set.
    z = snodes.back()->get_node_attr()->z;
  }
  delta; /*6*/
  //std::cout << "Ziterations: " << cnt << std::endl;
  delete ind->get_schedule();
  ind->set_schedule(new Schedule(*snodes.back()->get_schedule()));
  
  std::list<SNode*>::iterator sit = snodes.begin();
  for (; sit != snodes.end(); ++sit) {
    delete *sit;
    *sit = 0;
  }
  delete zset;
  zset = 0;
  
  /* Since we might have selected nodes that previously wasn't in
     zorder, we need to rebuild zorder and all priorities. */
  build_zorder(ind);
  build_instruction_priorities(ind);
#ifdef _SP_
#ifdef _TRANS_
  build_transfer_priorities(ind);
#endif
#endif
  delta; /*7*/
  // TEST
  //std::cout << *ind << std::endl;
  //std::cout << "--- Valid Schedule Created  ---" << std::endl;
  //exit(0);
  
  //  unlockit;
  delete zset;
  return ind->get_schedule();
}

void Population::match_node(instruction_candidates_t& candidates, vertex_id u_id,
                            SNode* current_snode, DAG& _dag, std::list<SNode*>& _snodes,
                            GNode* _gnode, Individual* _ind)
{
  //For all match y covering (u_id, current_snode, & chi, & instr)
  /* mater: get our own matcher (because of parallelization) */
  Matcher pmatcher(*matcher);
  Matcher::match_map_iter mi, mi_end;
  for(tie(mi, mi_end) = pmatcher.matches(u_id, *current_snode, _dag);
      mi != mi_end; ++mi) {
    std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;
    
    //condition checking
    int idx = 0;
    bool cond_check;
    cond_check = true;  //instruction without condition (default true)
    
    if((idx = mi->first->get_cond_index()) != -1) {
      cond_check = false; //instruction with condition (default false)
      //Apply condition check
      pattern_t *pattern = p.first;
      int psize = boost::num_vertices(*pattern);	 //pattern size
      std::vector< lcc::Node > arguments(psize); //arguments(size of the pattern)
      std::vector< iso_map_s > iso_maps = p.second;
      int iso_vect_size = iso_maps.size();
      for(int vi = 0; vi < iso_vect_size; vi++) {
        //ISO MAP
        node_map_t iso_mapp = iso_maps[vi].iso_map;
        
        node_map_t::iterator bi = iso_mapp.begin(), ei = iso_mapp.end();
        for(; bi != ei; ++bi) {
          //build the vector arguments[vi] = iso_map[vi]
          vertex_id vid = bi->second; //DAG vertex
          pvertex_id pvid = bi->first; //Pattern vertex
          arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
        }
        
        //ARG MAP
        node_map_t arg_mapp = iso_maps[vi].arg_map; 		//pattern arg_map (v -> lcc::Node)
        bi = arg_mapp.begin();
        ei = arg_mapp.end();
        for(; bi != ei; ++bi) {
          //build the vector arguments[vi] = iso_map[vi]
          vertex_id vid = bi->second; //DAG vertex
          pvertex_id pvid = bi->first; //pattern vertex
          arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
        }
        
        cond_check = (conditions[idx])(arguments);
	if(cond_check) {
	  break; //Exit the condition checking (we have a candidate)
        }
      }
    }
    if(!cond_check) {
      continue; //next candidate
    }
    //End of condition checking (make a method: FIXME)
    
    for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
        iso_iter != p.second.end(); ++iso_iter) {
      vertex_set_t chi;
      chi.clear();
      for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
          vi != (iso_iter->iso_map).end(); ++vi) {
        chi.insert(vi->second);
      }
      
      // oskar: Add the candidate:
      candidates.push_back(InstructionCandidate(mi->first, NORMAL_INSTRUCTION, chi,
                                                *iso_iter, p.first));
    }
  } //ends for all match y covering...
} /* Ends match_node */
  


instruction_candidates_t::iterator 
Population::find_transfer_candidate(instruction_candidates_t& candidates,
				    Individual* ind, vertex_id last_selected,
				    unsigned int num_previous_transfers)
{
  instruction_candidates_t::iterator selected_candidate_it = candidates.end();

#ifdef _SP_
#ifdef _TRANS_
  // Try to find a transfer priority
  //std::cout << "Looking for a transfer priority from: " << last_selected << std::endl;
  transfer_priority_t::iterator transit = ind->get_transfer_priorities().find(last_selected);
  if (transit != ind->get_transfer_priorities().end()) {
    std::vector<TargetInstruction*>& priorities = (*transit).second;
    //std::cout << "Found an array for " << last_selected << ", size: " << priorities.size()
    //          << ", num_prev_trans: " << num_previous_transfers << std::endl;
    if (priorities.size() > num_previous_transfers &&
        priorities[num_previous_transfers] != 0) {
      //std::cout << "Found a transfer priority vector." << std::endl;
      // There's a transfer priority. Check if we have a matching candidate.
      instruction_candidates_t::iterator inst_it = candidates.begin();
      for (; inst_it != candidates.end(); ++inst_it) {
        //std::cout << "  Looking for " << priorities[num_previous_transfers]->get_mnemonic()
        //          << ", checked " << (*inst_it).tinst->get_mnemonic() << std::endl;
        if (priorities[num_previous_transfers] == (*inst_it).tinst) {
          //std::cout << "Found transfer priority for " << last_selected
          //          << " to " << (*inst_it).tinst->get_mnemonic() << std::endl;
          selected_candidate_it = inst_it;
          break;
        }
      }
    }
  }
#endif
#endif
  
  return selected_candidate_it;
}

instruction_candidates_t::iterator 
Population::find_instruction_candidate(instruction_candidates_t& candidates,
				       Individual* ind, vertex_id u_id)
{
  instruction_candidates_t::iterator selected_candidate_it = candidates.end();
  
  // No transfer priority found, look for normal instruction priority.
  //std::cout << "Looking for instruction priorities for " << u_id << std::endl;
  instruction_priority_t::iterator prit = ind->get_instruction_priorities().find(u_id);
  if (prit != ind->get_instruction_priorities().end()) {
    instruction_candidates_t::iterator inst_it = candidates.begin();
    for (; inst_it != candidates.end(); ++inst_it) {
      //std::cout << "  Looking for " << (*prit).second->get_mnemonic()
      //          << ", checked " << (*inst_it).tinst->get_mnemonic() << std::endl;
      if ((*inst_it).tinst == (*prit).second) {
        //std::cout << "Found instruction priority for " << u_id
        //          << " to " << (*inst_it).tinst->get_mnemonic() << std::endl;
        selected_candidate_it = inst_it;
        break;
      }
    }
  }
  
  return selected_candidate_it;
}

InstructionType 
Population::select_instruction(instruction_candidates_t& candidates, vertex_id u_id,
			       vertex_id last_selected, SNode* current_snode,
			       std::list<SNode*>& snodes, GNode* gnode,
			       Individual* ind, unsigned int num_previous_transfers,
			       ZSet * zset)
{ 
  int MAX_ADVANCE = 10;
  int numtries = 0;
  std::vector<InstructionCandidate> second_try_candidates;
  
  bool candidate_selected = false;
  instruction_candidates_t::iterator selected_candidate_it = candidates.end();
  
  
  while (!candidate_selected) {
    if (candidates.size() > 0) {
      // Try to find a transfer priority
      selected_candidate_it = find_transfer_candidate(candidates, ind, last_selected, num_previous_transfers);
      
      if (selected_candidate_it == candidates.end()) {
        // No transfer priority found, look for normal instruction priority.
        selected_candidate_it = find_instruction_candidate(candidates, ind, u_id);
      }
      
      if (selected_candidate_it == candidates.end()) {
        /* No priority found, or the priority candidate isn't
           availible. Select a random and set a new priority. */
        int n = random_interval(0, candidates.size());
        selected_candidate_it = candidates.begin() + n;
      }
      
      candidate_selected = update_space(current_snode, *selected_candidate_it, snodes, gnode, zset);
      
      if (!candidate_selected) {
        /* This assumes that the false value is because a transfer
           can't be issued at this time, because the value it wants to
           transfer isn't ready. Other faults may be hidden because of
           this. See update_space() too. */
        second_try_candidates.push_back(*selected_candidate_it);
        candidates.erase(selected_candidate_it);
      } else {
        // Set transfer or instruction priority depending on what we choose!
        if ((*selected_candidate_it).type == NORMAL_INSTRUCTION){ 
	  if (ind->get_instruction_priorities().find(u_id) == ind->get_instruction_priorities().end()){
	    ind->get_instruction_priorities().insert(std::make_pair(u_id, (*selected_candidate_it).tinst));
	  }
#ifdef _SP_
#ifdef _TRANS_
        } else if ((*selected_candidate_it).type == TRANSFER_INSTRUCTION) {
          transfer_priority_t::iterator transit = ind->get_transfer_priorities().find(last_selected);
          if (transit == ind->get_transfer_priorities().end()) {
            transit = ind->get_transfer_priorities().
              insert(std::make_pair(last_selected, std::vector<TargetInstruction*>())).first;
          }
          std::vector<TargetInstruction*>& priorities = (*transit).second;
          //std::cout << "Filling from size(): " << priorities.size() << " to num_prev_trans: "
          //          << num_previous_transfers << std::endl;
          
          priorities.reserve(num_previous_transfers + 1); // To avoid new memory allocation in the loop
          for (unsigned int i=priorities.size(); i < num_previous_transfers; ++i) {
            //std::cout << "  filling nr " << i << std::endl;
            priorities.push_back(0); // Make sure we fill any empty slots.
          }
          //std::cout << "Done filling." << std::endl;
//           std::cout << "size before: " << priorities.size()
//                     << ", num_prev: " << num_previous_transfers << std::endl;
          priorities.push_back((*selected_candidate_it).tinst);
//           std::cout << "size after: " << priorities.size() << std::endl;
//           std::cout << "priorities[num_previous_transfers] = "
//                     << priorities[num_previous_transfers] << " ("
//                     << priorities[num_previous_transfers]->get_mnemonic() << ")" << std::endl;
          
          //std::cout << "Setting transfer priority from " << (*transit).first << " to "
          //          << (*selected_candidate_it).tinst->get_mnemonic() << std::endl;
#endif
#endif
        } else {
          std::cerr << "(GA::select_instruction) Error: Unknown instruction type!" << std::endl;
          exit(-1);
        }
      }
    } else {
      /*This code requires SpaceProfiles to work well. It *might* work
        without, but there's no guarantee that it will. */
      
#ifdef _SP_
      bool resources_full = true;
      int i = 0;
      std::vector< std::stack< int > >::const_iterator it =
        current_snode->get_space_profile()->get_resources().begin();
      for (; it != current_snode->get_space_profile()->get_resources().end(); ++it) {
        //std::cout << "Stack [" << i << "] size: " << (*it).size() << std::endl;
        if (!(*it).empty()) {
          resources_full = false;
        }
        ++i;
      }
      //std::cout << "Resources full? " << resources_full << std::endl;
      
      
      if (!resources_full && (numtries < MAX_ADVANCE)) {
        // Do a second try by advancing the schedule and timeprofile
        current_snode->get_time_profile()->advance(1);
        current_snode->get_schedule()->advance(1);
        current_snode->set_reference_time(current_snode->get_reference_time() + 1);
        
        candidates.insert(candidates.end(), second_try_candidates.begin(), second_try_candidates.end());
        second_try_candidates.clear();
        
        // TODO: Limit how many tries we do?
        ++numtries;
      } else {
        return NO_INSTRUCTION;
      }
#endif
    }
  }
  
  return (*selected_candidate_it).type;
}

/* The name is a bit misleading, there is no updating of any space */
bool Population::update_space(SNode* current_snode, InstructionCandidate candidate,
                              std::list<SNode*>& snodes, GNode* gnode, ZSet * zset)
{
  //static int times = 0;
  //std::cout << "(GA::update_space " << ++times << ")" << std::endl;
  
  NodeAttr* succ_node_attr = select(candidate.chi, *current_snode->get_node_attr(), dag, zset);
  SNode* succ_snode = new SNode(*current_snode);
  succ_snode->set_node_attr(succ_node_attr);
  
  std::pair< bool, int > ret = 
    succ_snode->insert(candidate.chi, dag, candidate.iso_map, *candidate.pattern,
                       candidate.tinst, architecture);
  
  if (!ret.first) {
    /* This assumes that the false value is because a transfer can't
       be issued at this time, because the value it wants to transfer
       isn't ready. Other faults may be hidden because of this. */
//     std::cerr << "(GA::update_space) Warning, insert returned false. Making a new choice." << std::endl;
    
//     Emitter* emitter = new Emitter(architecture);
//     emitter->emit(succ_snode->get_schedule());
//     delete emitter;
    
    delete succ_snode;
    succ_snode = 0;
    return false;
  }
  
  gnode->set_z(succ_snode->get_node_attr()->z);
  gnode->set_reference_time(succ_snode->get_reference_time());
  gnode->set_tau(succ_snode->get_tau());
  gnode->set_target_instruction(candidate.tinst);
//   std::cout << "TP: " << (int*)succ_snode->get_time_profile() << std::endl;
//   std::cout << *succ_snode->get_time_profile() << std::endl;
  gnode->set_time_profile(new TimeProfile(*(succ_snode->get_time_profile())));
#ifdef _SP_
  gnode->set_space_profile(new SpaceProfile(*(succ_snode->get_space_profile())));
#endif //_SP_
  gnode->set_instruction_type(candidate.type);
  
  snodes.push_back(succ_snode);
  
  return true;
}

NodeAttr* Population::select(const vertex_set_t & _chi, const NodeAttr & _node_attr, 
			     DAG & _dag, ZSet * zset)
{
  vertex_set_t z(_node_attr.z);
  indeg_t indeg(_node_attr.indeg);
  vertex_set_t alive;
  vertex_set_t::const_iterator chi_iter = _chi.begin();
  for(; chi_iter != _chi.end(); ++chi_iter) {
    vertex_id u_id = *chi_iter;
    //successors of v_id are decremented
    successor_iterator pi, pi_end;
    for(tie(pi, pi_end) = successors(u_id, _dag);
        pi != pi_end; ++pi) {
      vertex_id v_id = *pi;
      indeg[v_id]--;
      if(indeg[v_id] == 0) {
        z.insert(v_id);
      }
    }
  }
  //Remove chi nodes from z
  for(chi_iter = _chi.begin(); chi_iter != _chi.end(); ++chi_iter) {
    assert(indeg[*chi_iter] == 0);
    z.erase(*chi_iter);
  }
  //Alive set
  graph_traits< DAG >::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id u_id = source(*ei, _dag),
      v_id = target(*ei, _dag);
    if((indeg[u_id] == 0) && (z.find(u_id) == z.end())) {
      if((indeg[v_id] != 0) || (z.find(v_id) != z.end())) {
        //Target in zero-indegree set
        alive.insert(u_id);
      }
    }			
  }
  
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
}

NodeAttr* Population::get_init_node_attr(DAG& _dag, ZSet * zset)
{
  //oskar: This is from opts.cc
  graph_traits< DAG >::edge_iterator ei, ei_end;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  indeg_t indeg;
  vertex_set_t z;

  //All vertices start with indegree 0
  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    indeg[*vi] = 0;
  }
  //Initial indegrees
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id v_id = target(*ei, _dag);
    indeg[v_id]++;
  }	
  //Initial z
  indeg_t::iterator ii;
  for(ii = indeg.begin(); ii != indeg.end(); ++ii) {
    if(indeg[ii->first] == 0) {
      z.insert(ii->first);
    }
  }

  vertex_set_t alive;
  
  //Initial set of alive is empty if we only consider basic block level
  alive.clear();
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
}

void Population::transfers(instruction_candidates_t& candidates, SNode* current_snode,
                           std::list<SNode*>& _snodes, GNode* _gnode, Individual* _ind)
{
#ifdef _SP_
#ifdef _TRANS_
  static pattern_t dummy;
  //Transfers
  SpaceProfile current_sp = *(current_snode->get_space_profile());
  vertex_set_t alive = current_sp.get_alive();
  
  for(vertex_set_t::iterator a_iter = alive.begin(); 
      a_iter != alive.end(); ++a_iter) {
    vertex_id u = *a_iter;
    std::set< int > & rcs = current_sp[u];
    for(std::set< int >::iterator src_iter = rcs.begin(); src_iter != rcs.end();
        ++src_iter) {
      std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > > dests =
        architecture->get_transfer_destinations(*src_iter);
      if(!dests.first) {
        continue;
      }
      std::vector< std::pair< int, TargetInstruction * > >::iterator dti_iter = 
        dests.second.begin(), dti_iter_end = dests.second.end();
      for( ; dti_iter != dti_iter_end; ++dti_iter) {
        if(rcs.find(dti_iter->first) == rcs.end()) {
          //node u is not in residence class of the destination of the transfer instruction
          //add transfer instruction
          iso_map_s tmp_iso_maps;
          tmp_iso_maps.root = u;
          
          // oskar: Add candidate, mater: memory leak new pattern_t()!
          candidates.push_back(InstructionCandidate(dti_iter->second, TRANSFER_INSTRUCTION,
                                                    vertex_set_t(), tmp_iso_maps, &dummy));//new pattern_t()));
        }
      }
    }
  }

#endif //_TRANS_
#endif //_SP_
}


};
