#include "opts.hh"
#include <stdio.h>
#include <assert.h>
#include <algorithm>
#include <iomanip>
#include <stack>
#include "verbose.hh"
#include "cg.hh"

#define max(a,b) (((a)>(b))?(a):(b))
#define PRINT_SCHEDULES

//Equality and hashing for the grid
#ifdef __GNUC__
namespace __gnu_cxx 
#else
namespace std
#endif //__GNUC__
{
  template<>
  struct hash< SNode * >
  {
    size_t operator()(const SNode * s) const
    {
      return s->get_hash();
    }
  };
}; //std or __gnu_cxx
namespace std {
  template<> 
  struct equal_to< SNode * >
  {
    bool operator()(const SNode * s1, const SNode * s2) const
    {
      return *s1 == *s2;
    }
  };
}; //std

/*******************************************************************************
 *                        Optimist::Optimist()                                 *
 *******************************************************************************/
Optimist::Optimist(Architecture *_architecture)
  : architecture(_architecture), last_in(GRID_DIM)
{
  matcher = new Matcher(architecture);
  conditions = _architecture->get_conditions();
}

/*******************************************************************************
 *                        Optimist::~Optimist()                                *
 *******************************************************************************/
Optimist::~Optimist()
{
  delete matcher;
}

void Optimist::deallocate_nodes(solution_space_t & solution_space,
				int num_vertices, unsigned int t)
{
  //Memory management
  if((t > 0) && (t - 1 > 0)) { //memory criteria
    coord_t clear_v(GRID_DIM);
    clear_v[T_AXIS] = t - 1;
    std::stack< SNode * > tmp_stack; //hold pointer for deletion
    for(unsigned int l = 0; l < num_vertices; l++) {
      clear_v[L_AXIS] = l;
      grid_elem_t::iterator ci, ci_end;
      assert(tmp_stack.empty());
      for(tie(ci, ci_end) = solution_space.elements(clear_v); 
	  ci != ci_end; ++ci) {
	tmp_stack.push(*ci);
      }
      solution_space.clear(clear_v);
      while(!tmp_stack.empty()) {
	SNode *sn = tmp_stack.top();
	tmp_stack.pop();
	delete sn;
	sn = NULL;
      }
    }
  }
}

void Optimist::match_node(vertex_id u_id, SNode * current_snode, 
			  DAG & _dag, coord_t & coord, 
			  solution_space_t & solution_space) {
  Matcher::match_map_iter mi, mi_end;
  //For all match y covering (u_id, current_snode, & chi, & instr)
  for(tie(mi, mi_end) = matcher->matches(u_id, *current_snode, _dag);
      mi != mi_end; ++mi) {
    std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;

    //condition checking
    int idx = 0;
    bool cond_check;
    cond_check = true;  //instruction without condition (default true)

    if((idx = mi->first->get_cond_index()) != -1) {
      cond_check = false; //instruction with condition (default false)
      //Apply condition check
      pattern_t *pattern = p.first;
      int psize = boost::num_vertices(*pattern);	 //pattern size
      std::vector< lcc::Node > arguments(psize); //arguments(size of the pattern)
      std::vector< iso_map_s > iso_maps = p.second;
      int iso_vect_size = iso_maps.size();
      for(int vi = 0; vi < iso_vect_size; vi++) {
	//ISO MAP
	node_map_t iso_mapp = iso_maps[vi].iso_map;

	node_map_t::iterator bi = iso_mapp.begin(), ei = iso_mapp.end();
	for(; bi != ei; ++bi) {
	  //build the vector arguments[vi] = iso_map[vi]
	  vertex_id vid = bi->second; //DAG vertex
	  pvertex_id pvid = bi->first; //Pattern vertex
	  arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
	}

	//ARG MAP
	node_map_t arg_mapp = iso_maps[vi].arg_map; 		//pattern arg_map (v -> lcc::Node)
	bi = arg_mapp.begin();
	ei = arg_mapp.end();
	for(; bi != ei; ++bi) {
	  //build the vector arguments[vi] = iso_map[vi]
	  vertex_id vid = bi->second; //DAG vertex
	  pvertex_id pvid = bi->first; //pattern vertex
	  arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
	}

	cond_check = (conditions[idx])(arguments);
	if(cond_check) {
	  break; //Exit the condition checking (we have a candidate)
	}
      }
    }
    if(!cond_check) {
      continue; //next candidate
    }

    //End of condition checking (make a method: FIXME)

    for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
	iso_iter != p.second.end(); ++iso_iter) {
      vertex_set_t chi;
      chi.clear();
      for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
	  vi != (iso_iter->iso_map).end(); ++vi) {
	chi.insert(vi->second);
      }

      update_space(current_snode, mi->first, chi, coord, solution_space, 
		   *iso_iter,
		   *(p.first), _dag);
    }
  } //ends for all match y covering...
}

void Optimist::transfers(Flags & _flags, SNode * current_snode, DAG & _dag,
                         coord_t & coord, solution_space_t & solution_space)
{
#ifdef _SP_
#ifdef _TRANS_
  //Transfers
  SpaceProfile current_sp = *(current_snode->get_space_profile());
  vertex_set_t alive = current_sp.get_alive();

  int ts = _flags.i_ts;
  /* for each alive value */
  for(vertex_set_t::iterator a_iter = alive.begin(); 
      a_iter != alive.end(); ++a_iter) {
    vertex_id u = *a_iter;
    std::set< int > & rcs = current_sp[u];
    /* for each residence class */
    for(std::set< int >::iterator src_iter = rcs.begin(); src_iter != rcs.end();
	++src_iter) {
      /* for each possible target */
      /*       possible                       dest    move instr                   */
      std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > > dests =
	architecture->get_transfer_destinations(*src_iter);
      if(!dests.first) {
	continue;
      }
      /* for each possible transfer */
      std::vector< std::pair< int, TargetInstruction * > >::iterator dti_iter = 
	dests.second.begin(), dti_iter_end = dests.second.end();
      for( ; dti_iter != dti_iter_end; ++dti_iter) {
	if(rcs.find(dti_iter->first) == rcs.end()) {
	  //node u is not in residence class of the destination of the transfer instruction
	  //add transfer instruction
	  iso_map_s tmp_iso_maps;
	  tmp_iso_maps.root = u;
	  pattern_t dummy_pattern;
	  update_space(current_snode, dti_iter->second, vertex_set_t(), coord, 
		       solution_space, tmp_iso_maps,
		       //u, dummy_node_map, 
		       dummy_pattern, _dag);
	  if(_flags.i_ts != 0) {
	    ts--;
	    if(ts == 0) {
	      goto end_transfer;
	    }
	  }
	} /* ends for each possible transfer */
      } /* ends for each possible target */
    } /* ends for each residence class */
  } /* ends for each alive value */
  end_transfer: {}
#endif //_TRANS_
#endif //_SP_
}

/******************************************************************************
 *                        Optimist::schedule()                                *
 *----------------------------------------------------------------------------*
 * Called from cg.cc.                                                         *
 * for each time                                                              *
 *  for each level                                                            *
 *   for each node in solution_space (t,l)                                    *
 *    for each node in the zero indegree set                                  *
 *     for each matching pattern                                              *
 *      update_space()                                                        *
 ******************************************************************************/
Schedule *Optimist::schedule(DAG & _dag, Flags & _flags)
{
  int tmp;
  //Statistics
  cut = improved = merged =  n_snodes = 0;
  int omega = architecture->get_omega();
  zset = new ZSet;
  NodeAttr *node_attr = this->get_init_node_attr(_dag);

  TimeProfile *init_time_profile = new TimeProfile(architecture->get_reserv_tab_size());
  SNode *init_snode =
    new SNode(node_attr, init_time_profile, architecture);

  //Reduce SNode: compute ADDR and CNST nodes to their respective residences
  int num_reduced = 0; 

  int num_vertices = boost::num_vertices(_dag);
  coord_t coord(GRID_DIM), stop_coord(GRID_DIM);

  //Initialization
  stop_coord[T_AXIS] = 0;

  coord[T_AXIS] = 0;
  coord[L_AXIS] = 0;
  max_coord = coord;
  solution_space_t solution_space;

  solution_space.insert(init_snode, coord);
  VERB1 << "num_vertices: " << num_vertices << "\n"; 
  VERB2 << "num_reduced: " << num_reduced << "\n"; 
  num_vertices -= num_reduced;
  VERB2 << "num_vertices (- num_reduced): " << num_vertices << "\n"; 
  stop_coord[L_AXIS] = num_vertices;

  //The driving loop
  //for each t
  for(unsigned int t = 0; t <= max_coord[T_AXIS]; t++) {
    nr_greedy.push_back(0);
    nr_str_lin.push_back(0);
#ifndef _NOMM_  //default in makefile: not defined.
    /* solution_space is passed as a reference */
    deallocate_nodes(solution_space, num_vertices, t);
#endif //_NOMM_

    //check stop
    stop_coord[T_AXIS] = t;
    /* Break on entry? */
    if(!solution_space.empty(stop_coord)) {
      std::cout << "\nsolution found (on entry): tau = " << t //<< std::endl
		<< "\n n_snodes: " << n_snodes
		<< " merged: " << merged
		<< " cut: " << cut
		<< " improved: " << improved << "\n"; //std::endl;
      break;
    }

    //for each level
    for(unsigned int l = 0; l < num_vertices; l++) {
      VERB2 << "                           \r";
      VERB2 << std::setw(3) << "l: " << std::setw(3) << l 
	    << "/" << std::setw(3) << num_vertices 
	    << std::setw(3) << " t:" << std::setw(3) << t
	    << std::setw(4) << " #n:" << std::setw(6) << n_snodes;

      coord[T_AXIS] = t;
      coord[L_AXIS] = l;
      grid_elem_t::iterator gi, gi_end;
      //for each node in solutionspace (t,l)
      for(tie(gi, gi_end) = solution_space.elements(coord);
	  gi != gi_end; ++gi) {
	SNode *current_snode =  *gi;

	vertex_set_t::iterator zi, zi_end;
	NodeAttr *current_node_attr = current_snode->get_node_attr();
	zi = current_node_attr->z.begin();
	zi_end = current_node_attr->z.end();

	//Heuristic
	int hs = _flags.i_hs;

	//for each node in z
	for( ; zi != zi_end; ++zi) {
	  vertex_id u_id = *zi;
	  /* _dag, coord and solution_space are passed as references. */
	  match_node(u_id, current_snode, _dag, coord, solution_space);

	  //Heuristic: number of variants
	  if(_flags.i_hs != 0) {
	    hs--;
	    if(hs == 0) {
	      break;
	    }
	  }
	} //ends for each node in z
        transfers(_flags, current_snode, _dag, coord, solution_space);
      } //ends for each node in solutionspace (t,l)
      tmp = l;
    } //ends for each level
    //In case the last node does not increase the total execution time
    if(!solution_space.empty(stop_coord)) {
#if 0 /* For the cpc paper. When this was active we don't break on
	 found solution.! */
      classify_solutions(stop_coord, _dag, solution_space);
#endif
       VERB1 << "\n solution found (on exit): tau = " << t
 	    << "\n n_snodes= " << n_snodes
 	    << "\n merged= " << merged
 	    << "\n cut= " << cut
 	    << "\n improved= " << improved
 	    << "\n l = " << tmp << "\n";
       break;
    }
  } //ends for each t

  grid_elem_t::iterator gi, gi_end;
  tie(gi, gi_end) = solution_space.elements(stop_coord);
  if(gi == gi_end) {
    //No solution found
    VERB1 << "\nn_snodes: " << n_snodes
	  << " merged: " << merged
	  << " cut: " << cut
	  << " improved: " << improved << "\n";

    assert(!solution_space.empty(last_in));

    std::cerr << "Solution not found!\n";
    exit(-1);
  }
  Schedule *s = (*gi)->get_schedule();

  std::cout << "Optimist::schedule() exits: \n n_snodes= " << n_snodes
	    << "\n merged= " << merged
	    << "\n cut= " << cut
	    << "\n improved= " << improved 
	    << "\n l= " << tmp << "\n";

  delete init_time_profile;
  delete init_snode;
  delete zset;
  //  s->print_reserv_tab(); //mater test;
  return s;
} //ends Optimist::schedule()

/*******************************************************************************
 *                        Optimist::update_space()                             *
 * succ_snode = copy(curr_snode);                                              *
 * succ_snode->insert(new stuff);                                              *
 * either                                                                      *
 *  discard succ_snode;                                                        *
 * or                                                                          *
 *  solution_space->insert(succ_snode);                                        *
 *******************************************************************************/
void Optimist::update_space(const SNode *_current_snode,
			    TargetInstruction *_target_instruction,
			    const vertex_set_t & _chi, //mater: covered nodes
			    const coord_t & _current_coord,
			    solution_space_t & _solution_space,
			    iso_map_s & _iso_maps,
			    pattern_t & _pattern,
			    DAG _dag)
{
  NodeAttr *succ_node_attr = 
    select(_chi, *_current_snode->get_node_attr(), _dag);
  SNode *succ_snode = new SNode(*_current_snode);
  succ_snode->set_node_attr(succ_node_attr);

  std::pair< bool, int > p = 
    succ_snode->insert(_chi,
		       _dag, _iso_maps,
		       _pattern, _target_instruction, architecture);

  if(!p.first) {
    delete succ_snode;
    cut++;
    return;
  }

  coord_t check_coord(GRID_DIM), dest_coord(GRID_DIM);
  check_coord = _current_coord;

  int tau = succ_snode->get_tau();
  //FIXME  assert(tau > _current_coord[T_AXIS]);
  int dest_l = _current_coord[L_AXIS] + _chi.size();
  dest_coord[L_AXIS] = dest_l;
  dest_coord[T_AXIS] = tau;

  check_coord[L_AXIS] = dest_l;
  check_coord[T_AXIS] = _current_coord[T_AXIS];
  SNode *snode_prime = NULL;
  unsigned int maxtime = max_coord[T_AXIS];

  /* mater: Check if a solution, s == succ_node, already exists. */
  std::pair< grid_elem_t::iterator, bool > ep;
  for( ; check_coord[T_AXIS] <= maxtime; check_coord[T_AXIS]++) {
    ep = _solution_space.exist(succ_snode, check_coord);
    if(ep.second) {
      break;
    }
  }
  if(ep.second) { /* mater: a solution comparable to succ_snode existed. */
    /* Idea: If we do a classification, we can not allow snodes to be
       removed because they do not have the same time. */
    SNode *tmp_snode = *ep.first;
    assert(*tmp_snode == *succ_snode);
    if(tau < check_coord[T_AXIS]) {
      //Improved intermediate solution
      _solution_space.erase(*(ep.first), check_coord);
      SNode *sn = *(ep.first);
      delete sn;
      sn = NULL;
      _solution_space.insert(succ_snode, dest_coord);
      improved++;
    } else {
      //New solution is equivalent or worse (forget it)
      /* Idea: If we do merging when classifying this succ_node may be
	 deleted and we must update a count variable in the existing
	 (comparable) snode.*/
      /* We know for sure that both of the schedules are non-greedy OR
	 they are both greedy and represent the same schedule.
	 (because we edited snode::operator==). */
      /* If the two schedules are not equal, increase the counter. */
      if(*(tmp_snode->get_schedule()) != *(succ_snode->get_schedule()))
	tmp_snode->increase_merged(succ_snode->get_merged());
      delete succ_snode;
      merged++;
    }
  } else {
    //New intermediate solution
    ep = _solution_space.insert(succ_snode, dest_coord);
    last_in = dest_coord;
    assert(ep.second); //Insertion succeeded
    n_snodes++;
    //Update max_coord
    max_coord[T_AXIS] = max(max_coord[T_AXIS], tau);
  }
}

/*******************************************************************************
 *                     Optimist::get_init_node_attr()                          *
 *******************************************************************************/
NodeAttr *Optimist::get_init_node_attr(DAG & _dag)
{
  graph_traits< DAG >::edge_iterator ei, ei_end;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  indeg_t indeg;
  vertex_set_t z;
	
  //All vertices start with indegree 0
  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    indeg[*vi] = 0;
  }
  //Initial indegrees
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id v_id = target(*ei, _dag);
    indeg[v_id]++;
  }	
  //Initial z
  indeg_t::iterator ii;
  for(ii = indeg.begin(); ii != indeg.end(); ++ii) {
    if(indeg[ii->first] == 0) {
      z.insert(ii->first);
    }
  }
	
  vertex_set_t alive;
	
  //Initial set of alive is empty if we only consider basic block level
  alive.clear();
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
	
}

/*******************************************************************************
 *                        Optimist::select()                                   *
 * Nodes in _chi are selected and the node attributes are updated.             *
 *******************************************************************************/
NodeAttr *Optimist::select(const vertex_set_t & _chi, const NodeAttr & _node_attr,
			   DAG & _dag)
{
  vertex_set_t z(_node_attr.z);
  indeg_t indeg(_node_attr.indeg);
  vertex_set_t alive;
  vertex_set_t::const_iterator chi_iter = _chi.begin();
  for(; chi_iter != _chi.end(); ++chi_iter) {
    vertex_id u_id = *chi_iter;
    //successors of v_id are decremented
    successor_iterator pi, pi_end;
    for(tie(pi, pi_end) = successors(u_id, _dag);
	pi != pi_end; ++pi) {
      vertex_id v_id = *pi;
      indeg[v_id]--;
      if(indeg[v_id] == 0) {
	z.insert(v_id);
      }
    }
  }
  //Remove chi nodes from z
  for(chi_iter = _chi.begin(); chi_iter != _chi.end(); ++chi_iter) {
    assert(indeg[*chi_iter] == 0);
    z.erase(*chi_iter);
  }
  //Alive set
  graph_traits< DAG >::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id u_id = source(*ei, _dag),
      v_id = target(*ei, _dag);
    if((indeg[u_id] == 0) && (z.find(u_id) == z.end())) {
      if((indeg[v_id] != 0) || (z.find(v_id) != z.end())) {
	//Target in zero-indegree set
	alive.insert(u_id);
      }
    }			
  }
	
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
}

/*******************************************************************************
 *                        Optimist::classify_solutions()                       *
 *******************************************************************************/
void Optimist::classify_solutions(coord_t _coord, DAG & _dag, solution_space_t & _solution_space)
{
  std::cout << "\nClassification of schedules, format is time, nr_greedy, nr_str_lin (n_snodes)\n";
  int n_snodes = 0;
  Schedule *ss;
  grid_elem_t::iterator gi,gi_end;
  tie(gi,gi_end) = _solution_space.elements(_coord);
  for(;gi!=gi_end;++gi){
#ifdef PRINT_SCHEDULES
    if((*gi)->greedy())
      std::cout << "GREEDY SCHEDULE:\n";
    else
      std::cout << "STR LIN SCHEDULE:\n";
    (*gi)->get_schedule()->print_resource_usage(architecture);
    Emitter *emit = CG::get_emitter();
    emit->emit((*gi)->get_schedule());
#endif
    if((*gi)->greedy())
      nr_greedy.back()++;
    else
      nr_str_lin.back()+=(*gi)->get_merged();
    n_snodes++;
    //     ss = (*gi)->get_schedule();
    //     if (ss->greedy(_dag,architecture))
    //       nr_greedy.back()++;
    //     else
    //       nr_str_lin.back()++;
  }
  std::cout << ">" << _coord[T_AXIS]
	    << " " << nr_greedy.back()
	    << " " << nr_str_lin.back() + nr_greedy.back()
	    << " " << n_snodes << std::endl;
}


/* mater: removed 1882828338 lines of commented out code */
