#include <algorithm>
#include <assert.h>
#include "sp.hh"

SpaceProfile::SpaceProfile(Architecture & _architecture)
{
  std::vector< std::vector < int > > & 
    _rc_resources = _architecture.get_rc_resources();
  for(int i = 0; i < _rc_resources.size(); i++) {
    resources.push_back( std::stack< int >() );
    assert(resources.size() == i + 1);
    if((_rc_resources[i])[0] == -1) {  //Unlimited size
      resources[i].push(-1);
    } else {
      for(int j = 0; j < _rc_resources[i].size(); j++) {
	resources[i].push(_rc_resources[i][j]);
      }
    }
  }
  this->clear();
  nodes.clear();
  instances_map.clear();
  alive.clear();
}

SpaceProfile::SpaceProfile(const SpaceProfile & _space_profile)
  : std::map< vertex_id, std::set< int > >(_space_profile)
{
  nodes = _space_profile.nodes;
  resources = _space_profile.resources;
  instances_map = _space_profile.instances_map;
  alive = _space_profile.alive;
}

SpaceProfile & SpaceProfile::operator=(const SpaceProfile & _space_profile)
{

  for(std::map< vertex_id, std::set < int > >::const_iterator si = 
	_space_profile.begin(); si != _space_profile.end(); ++si) {
    (*this)[si->first] = si->second;
  }
  nodes = _space_profile.nodes;
  resources = _space_profile.resources;
  instances_map = _space_profile.instances_map;
  alive = _space_profile.alive;
  return *this;
}

SpaceProfile::~SpaceProfile()
{
}

std::pair< bool, std::pair< char *, lcc::Node > >
SpaceProfile::insert(const vertex_set_t & _alive,
		     const vertex_id & _root, DAG & _dag,
		     int _available_at,
		     int _residence_class, Architecture *_architecture,
		     const std::pair< vertex_id, int > _dest1_is)
{
  //  vertex_set_t all_nodes;
  vertex_set_t dead;

  std::insert_iterator< vertex_set_t >
    out_ins(dead, dead.begin());

  std::set_difference(alive.begin(), alive.end(),
		      _alive.begin(), _alive.end(), out_ins);

  alive.clear();
  alive = _alive;

  if(_dest1_is.first != (vertex_id)-1) {
    assert(dead.find(_dest1_is.first) != dead.end());
  }

  instance_t saved_resource;

  for(vertex_set_t::iterator vi = dead.begin(); vi != dead.end(); ++vi) {
    //Return resources
    std::vector< std::pair< where_when_t, instance_t > > & tuple_v = instances_map[*vi];
    for(int i = 0; i < tuple_v.size(); i++) {
      std::pair< where_when_t, instance_t > & p = tuple_v[i];
      if(resources[p.first.first].empty() || resources[p.first.first].top() != -1) {
	// residence
	//If the resource being returned is reserved for destination, saved it, otherwise return
	if( (_dest1_is.first != (vertex_id)-1) && (_dest1_is.first == *vi) && (_dest1_is.second == p.first.first) ) {
	  saved_resource = p.second;
	} else {
	  resources[p.first.first].push(p.second.i);
	}
      }
    }
    //Remove node from the space profile and instances_map
    this->erase(*vi);
    instances_map.erase(*vi);
  }

  if(_dest1_is.first != (vertex_id)-1) {
    //We return previously allocated resource to one of the source nodes.
    saved_resource.p = get_lcc_node(_root, _dag);
    //FIXME , _available_at
    instances_map[_root].push_back(std::make_pair(std::make_pair(_residence_class,
								 _available_at), saved_resource));
    //    resources[_residence_class].pop();
    (*this)[_root].insert(_residence_class);
    char *reg_str = _architecture->get_reg_name(saved_resource.i);
    return std::make_pair(true, std::make_pair(reg_str, saved_resource.p));
  }

  //No resource need for STMTs
  if(_residence_class == STMT_RC) {
    return std::make_pair(true, std::make_pair((char *)"", (lcc::Node)NULL));
  }
  
  //If not enough resources
  if(resources[_residence_class].empty()) {
    return std::make_pair(false, std::make_pair((char *)"", (lcc::Node)NULL));
  }

  if(resources[_residence_class].top() != -1) {
    instance_t reg; 
    reg.i = resources[_residence_class].top();
    reg.p = get_lcc_node(_root, _dag);
    //FIXME , _available_at
    instances_map[_root].push_back(std::make_pair(std::make_pair(_residence_class,
								 _available_at), reg));
    resources[_residence_class].pop();
    (*this)[_root].insert(_residence_class);
    char *reg_str = _architecture->get_reg_name(reg.i);
    return std::make_pair(true, std::make_pair(reg_str, reg.p));
  } else {
    lcc::Node p = get_lcc_node(_root, _dag);

    switch(generic(p->op)) {

    case lcc::ADDRF:  case lcc::ADDRG: case lcc::ADDRL: case lcc::CNST:
    case lcc::LABEL:
      {
	instance_t symbol;
	//FIXME
	if(!(p->syms[0]->x.name) && (p->syms[0]->name)) {
	  p->syms[0]->x.name = p->syms[0]->name;
	}
	symbol.s = p->syms[0]->x.name;
	symbol.p = p;
	//FIXME , _available_at));
	instances_map[_root].push_back(std::make_pair(std::make_pair(_residence_class,
								     _available_at), symbol));
      }
      break;

    default:
      std::cout << std::endl << std::endl << p->op << std::endl << std::endl;
      assert(false);
    }
    (*this)[_root].insert(_residence_class);
    //FIXME
    if(!(p->syms[0]->x.name) && (p->syms[0]->name)) {
      p->syms[0]->x.name = p->syms[0]->name;
    }
    return std::make_pair(true, std::make_pair(p->syms[0]->x.name, p));
  }

  return std::make_pair(false, std::make_pair((char *)"", (lcc::Node)NULL));
}

std::pair< bool, std::pair< char *, lcc::Node > >
SpaceProfile::get_instance(vertex_id & _u, int _residence_class,
			   int _reference_time, Architecture *_architecture,
			   int *_earliest_instance)
{
  assert(*_earliest_instance == -1);
  assert(_residence_class < resources.size());
  assert(instances_map.find(_u) != instances_map.end());
  //  std::cout << "Lookup for: u: " << _u << " rc: " << _residence_class << " ref_time: " << _reference_time << std::endl;
  std::vector< std::pair< where_when_t, instance_t > > & inst = instances_map[_u];
  for(int i = 0; i < inst.size(); i++) {
    if(inst[i].first.first == _residence_class) {
      *_earliest_instance = inst[i].first.second;
      if(resources[_residence_class].empty() || 
	 resources[_residence_class].top() != -1) {
	return std::make_pair((*_earliest_instance <= _reference_time), 
		 std::make_pair(_architecture->get_reg_name(inst[i].second.i), 
				inst[i].second.p));
      }
      //Symbol
      return std::make_pair((*_earliest_instance <= _reference_time), std::make_pair(inst[i].second.s, inst[i].second.p));
    }
  }
  //  std::cout << "Failed lookup for: u: " << _u << " rc: " << _residence_class << " ref_time: " << _reference_time << std::endl;
  //  std::cout << (*this);
  return std::make_pair(false, std::make_pair((char *)0, (lcc::Node)NULL));
}

std::ostream & operator<<(std::ostream & _out, const SpaceProfile & _space_profile)
{
  std::map< vertex_id, std::set < int > >::const_iterator vmi;

  for(vmi = _space_profile.begin(); vmi != _space_profile.end(); ++vmi) {
    _out << "v: " << vmi->first << " rc: ";
    for(std::set< int >::const_iterator rci =  (vmi->second).begin();
	rci != (vmi->second).end(); ++rci) {
      _out << *rci << " ";
    }
    _out << std::endl;
  }
  //instances map
  std::map< vertex_id, std::vector< std::pair< SpaceProfile::where_when_t, SpaceProfile::instance_t > > >::const_iterator 
    im = _space_profile.instances_map.begin(), im_end = _space_profile.instances_map.end();
  for( ; im != im_end; ++im) {
    vertex_id u = im->first;
    std::vector< std::pair< SpaceProfile::where_when_t, SpaceProfile::instance_t > >::const_iterator 
      si = im->second.begin(), si_end = im->second.end();
    _out << "u: " << u << " in:\n";
    for( ; si != si_end; ++si) {
      int rc = si->first.first;
      int avail_time = si->first.second;

      int in_i = (si->second).i;
      char *in_str;
      if(in_i > 10)
	in_str = (si->second).s;

      _out << "rc: " << rc << " avail_time: " << avail_time;
      if(in_i > 10)
	_out << " inst.s: " << in_str << std::endl;
      else
	_out << " inst.i: " << in_i << std::endl;
    }
  }
  return _out;
}

/*


    for(std::map< int, std::set< vertex_set_t > >::iterator pi = pressure.begin();
	pi != pressure.end(); ++pi) {
      std::set< vertex_set_t > tmp = pi->second;
      for(std::set< vertex_set_t >::iterator si = tmp.begin(); si != tmp.end(); ++si) {
 	if((*si).find(*vi) != (*si).end()) {
 	  tmp.erase(si);
 	}
      }
    }
  }

  //Insert nodes
  for(vertex_set_t::const_iterator chi_i = _chi.begin(); chi_i != _chi.end(); ++chi_i) {
    ((*this)[*chi_i]).insert(_chi_residence_class);
  }
  nodes.insert(_chi);
  pressure[_chi_residence_class].insert(_chi);

//   std::map< vertex_id, int >::iterator vmi = this->begin();
//   for(vmi = this->begin(); vmi != this->end(); ++vmi) {
//     (pressure[vmi->second]).insert(vmi->first);
//   }

//   if(nodes != new_node_set) {
//     std::cerr << "\nDUMP\nnodes:\n";
//     std::copy(nodes.begin(), nodes.end(), std::ostream_iterator< vertex_id >(std::cerr, " "));
//     std::cerr << "\nnew_node_set:\n";
//     std::copy(new_node_set.begin(), new_node_set.end(), 
// 	      std::ostream_iterator< vertex_id >(std::cerr, " "));
//     std::cerr << "\n_alive:\n";
//     std::copy(_alive.begin(), _alive.end(), std::ostream_iterator< vertex_id >(std::cerr, " "));
//     std::cerr << "\n_under_computation:\n";
//     std::copy(_under_compuation.begin(), _under_compuation.end(), 
// 	      std::ostream_iterator< vertex_id >(std::cerr, " "));
//     std::cerr << "\n_chi:\n";
//     std::copy(_chi.begin(), _chi.end(), std::ostream_iterator< vertex_id >(std::cerr, " "));
//     std::cerr << "\nout_nodes:\n";
//     std::copy(out_nodes.begin(), out_nodes.end(), 
// 	      std::ostream_iterator< vertex_id >(std::cerr, " "));

//   }
//   assert(nodes == new_node_set);

}
*/

  //    new_ins(new_node_set, new_node_set.begin());
  /*
  std::set_union(_alive.begin(), _alive.end(),
		 _under_compuation.begin(), _under_compuation.end(),
		 new_ins);
  */
  //  all_nodes.clear();
  /*
  for(std::set< vertex_set_t >::iterator node_set = nodes.begin();
      node_set != nodes.end(); ++node_set) {
    vertex_set_t tmp = *node_set;
    for(vertex_set_t::iterator v = tmp.begin();	v != tmp.end(); ++v) {
      all_nodes.insert(*v);
    }
  }
  */

/*
SpaceProfile::SpaceProfile()
{
  this->clear();
  nodes.clear();
  pressure.clear();
  resources.clear();
}
*/

  //DEBUG
//   std::cerr << "\nDUMP\nall_nodes:\n";
//   std::copy(all_nodes.begin(), all_nodes.end(), 
// 	    std::ostream_iterator< vertex_id >(std::cerr, " "));
//   std::cerr << "\n_alive:\n";
//   std::copy(_alive.begin(), _alive.end(), std::ostream_iterator< vertex_id >(std::cerr, " "));
//   std::cerr << "\n_chi:\n";
//   std::copy(_chi.begin(), _chi.end(), std::ostream_iterator< vertex_id >(std::cerr, " "));
//   std::cerr << "\nout_nodes:\n";
//   std::copy(out_nodes.begin(), out_nodes.end(), 
// 	    std::ostream_iterator< vertex_id >(std::cerr, " "));

  //END DEBUG
