#include <assert.h>
#include "vcg.hh"

/**
 * Generates VCG graph for a forest of DAGs.
 */
VCG::VCG(lcc::Node p, std::ofstream & _f)
{
  header(_f);
  //Initialize nodes (tag them with a value)
  lcc::Node q = NULL;
  nodes.x.next = NULL;
  int n = 0;

  //Add nodes (each node has an id saved in x.inst (temp)
  for(n = 0, q = p; q; q = q->link) {
    n = add_node(q, n, _f);
  }
  //Add nodes and edges from the list of nodes
  for(q = nodes.x.next; q; q = q->x.next) {
    assert(q->x.inst != 0);
    _f << "node: {\n"
       << "  title: \"" << q->x.inst << "\"\n"
       << "  label: \"" << "(vid=" << q->x.inst << ")[" 
       << lcc::opname(q->op) << "]\"\n"
       << "  info1: \"" << q->x.inst << "\"\n"
       << "  info2: \"" << (void *)q << "\"\n"
       << "  info3: \"" << q->op << "\"\n"
       << "  color: white\n"
       << "}\n";

    for(int k = 0; k < 2; k++) {
      if(q->kids[k]) {
	assert(q->kids[k]->x.inst != 0 && q->x.inst != 0);
	_f << "edge: { sourcename: \"" << q->kids[k]->x.inst
	   << "\" targetname: \"" << q->x.inst
	   << "\" color: black "
	   << "label: \"" << "" << "\" }\n";
      }
    }
  }

  //Add links
  for(q = p; q; q = q->link) {
    if(q->link) {
      _f << "edge: { sourcename: \"" << q->x.inst
	 << "\" targetname: \"" << q->link->x.inst
	 << "\" color: green "
	 << "label: \"" << "" << "\" }\n";
    }
  }
  footer(_f);
}


VCG::VCG(DAG _dag, std::ofstream & _f)
{
  //node_property_map::type lcc_node_property(dag);
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  graph_traits< DAG >::edge_iterator ei, ei_end;
  lcc::Node p;
  header(_f);

  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    p = get_lcc_node(*vi, _dag);
    _f << "  node: {\n"
       << "    title: \"" << *vi << "\"\n"
       << "    label: \"" << "(vid=" << *vi << ")["
       << lcc::opname(p->op) << "]";
    if ( p->syms && p->syms[0] && p->syms[0]->scope == lcc::CONSTANTS ){
      /* A bug: also ASGN gets a value printed. */
      _f << "(" << p->syms[0]->u.value << ")";
    }
    if ( p->syms && p->syms[0] && p->syms[0]->scope != lcc::CONSTANTS ){
      _f << ":" << (p->syms[0]->name);
    }
    _f << "\"\n"
       << "    info1: \"" << *vi << "\"\n"
       << "    info2: \"" << (void *)p << "\"\n"
       << "    info3: \"" << p->op << "\"\n"
       << "    color: white\n"
       << "  }\n";
  }

  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    size_t p_id = source(*ei,_dag), q_id = target(*ei,_dag);
    _f << "  edge: { sourcename: \"" << p_id
       << "\" targetname: \"" << q_id
       << "\" color: black "
       << "label: \"" << edgetype((get_edge_type(*ei, _dag))) << "\" }\n";
  }
  footer(_f);
}

// p: current node, q: predecessor
void VCG::add_edge(lcc::Node p, lcc::Node q, std::ofstream & _f)
{
  if(p) {
    if(q) {
      _f << "edge: { sourcename: \"" << p->x.inst
	 << "\" targetname: \"" << q->x.inst
	 << "\" color: black "
	 << "label: \"" << "" << "\" }\n";

      add_edge(p->kids[0], p, _f);
      add_edge(p->kids[1], p, _f);
      //      p->x.inst = 0;
    } else {
      add_edge(p->kids[0], p, _f);
      add_edge(p->kids[1], p, _f);
    }

  }
}


int VCG::add_node(lcc::Node p, int n, std::ofstream & _f)
{
  if(p && p->x.inst == 0) {
    lcc::Node tail;
    p->x.inst = ++n;
    tail = nodes.x.next;
    nodes.x.next = p;
    p->x.next = tail;
    /*
    assert(p->x.inst != 0);
    _f << "node: {\n"
       << "  title: \"" << n << "\"\n"
       << "  label: \"" << "[" << lcc::opname(p->op) << "]\"\n"
       << "  info1: \"" << p->x.inst << "\"\n"
       << "  info2: \"" << "" << "\"\n"
       << "  color: white\n"
       << "}\n";
    */
    n = add_node(p->kids[0], n, _f);
    n = add_node(p->kids[1], n, _f);
  }
  return n;
}

void VCG::mark_node(lcc::Node p)
{
  if(p && p->x.inst != 0) {
    p->x.inst = 0;
    mark_node(p->kids[0]);
    mark_node(p->kids[1]);
  }
}

void VCG::header(std::ofstream &_f)
{ 
  // Header for VCG
  _f << "graph: { title:\"Graph\"\n"
     << "orientation: bottom_to_top\n"
     << "display_edge_labels: no\n"
     << "layoutalgorithm: minbackward\n"
     << "display_edge_labels: yes\n"
     << "dirty_edge_labels: no\n";
}

void VCG::footer(std::ofstream &_f)
{
  // Footer for VCG
  _f << "}\n";
}
